****************************************************************************
*
*		NOTE: All of these functions and procedures are free to used in 
*				any programs you wish without any fees or royalties.  I have 
*				tested these functions as thoroughly as possible and have found 
*				them to be bug-free.  If you come up with any enhancements or 
*				problems	please drop me a message.  Also, look for more uploads
*				in the future such as pushing and popping windows, a full-blown
*				data dictionary, and a reporting engine that automatically 
*				handles printing to the printer, the screen, or to any of the 
*				file formats offered in FoxPro's COPY TO command.
*
*
*		Thanks !
*
*		Tim Young		Compuserve	76570,607
*							MCI Mail 	LCCASA 
*
****************************************************************************

****************************************************************************
*																									 
*			C H A N G E S   M A D E   S I N C E   P U S H P O P . Z I P
*
*										File Date 3-19-93
*
*	->>>	Added trapping for eof() when pushing the recno(), thus avoiding 
*			trying to restore the recno() to eof() and generating an "Record 
*			out of bounds" error.																									
*	
*  ->>>	Added the ability to pop any database alias off the stack in any 	
*			order, which is ideal for a multi-window environment.
*
****************************************************************************

****************************************************************************
*																									 
*		Procedures:	PUSH_DBF 
*						POP_DBF [ WITH < alias name > ]
*																									
*		Purpose:		Saves all dbf information for the current work area 		
*						on an array-based stack to be retrieved at a later time
*						via the pop_dbf procedure.  Allows the programmer to	
*						create completely generic functions and routines and not 
*						have to be concerned about the current work area state 
*						and making sure that everything is restored after the 
*						function or routine completes.																			
*
*		Notes:		You *must* make sure that you declare a public array called
*						dbf_array (it doesn't matter what size it is initialized to
*						because the push_dbf procedure will re-size the array to 
*						match the stack pointer the first time it is called) and you
*						must also initialize a public variable called dbf_ptr to a
*						value of 0. It is *very* important that the dbf_ptr variable
*						is initialized to 0, any other value will not crash the 
*						program but will result in the unnecessary use of memory by
*						empty array rows.  Also please note, these functions assume
*						the use of structural index tags and do not accomodate .IDX
*						or non-structural .CDX files.
*
*		Sample:		FUNCTION get_reject
*						PARAMETERS chk_rej
*						PRIVATE result
*
*		This is 		IF lastkey()=27 .OR. readkey(1)=6
*		all there		RETURN .T.
*		is to it !	ENDIF
*
*		>>>>>>>>>>	DO push_dbf
*
*						result=.F.
*
*						SELECT reject
*						SET ORDER TO TAG code
*
*						IF seek(chk_rej)
*		>>>>>>>>>>		DO pop_dbf
*							RETURN .T.
*						ELSE
*							DO done_tone
*							IF eof()
*								GO TOP
*							ENDIF
*						ENDIF
*
*						DEFINE POPUP rej_pop ;
*								 FROM 6,15 ;
*								 TO 14,60 ;
*								 IN SCREEN ;
*								 MARGIN ;
*								 PROMPT FIELD reject.code+"  "+reject.reason ;
*								 SHADOW ;
*								 COLOR SCHEME 14
*
*						ON SELECTION POPUP rej_pop DO check_pop WITH lastkey(),"rej_pop"
*
*						ON KEY LABEL RIGHTARROW *
*						ON KEY LABEL LEFTARROW *
*		 
*						ACTIVATE POPUP rej_pop REST
*
*						ON KEY LABEL RIGHTARROW 
*						ON KEY LABEL LEFTARROW 
*
*						IF lastkey() = 13
*
*							result=.T.
*							chk_rej=reject.code   
*
*						ENDIF
*
*						SHOW GETS
*
*						DEACTIVATE POPUP rej_pop
*						RELEASE POPUP rej_pop
*
*		>>>>>>>>>>	DO pop_dbf
*
*						RETURN result
*																									
****************************************************************************
PROCEDURE push_dbf

dbf_ptr=dbf_ptr+1

DECLARE dbf_array[dbf_ptr,24]

dbf_array[dbf_ptr,1]=alias()
dbf_array[dbf_ptr,2]=order()
dbf_array[dbf_ptr,3]=iif(eof(),reccount(),recno())
dbf_array[dbf_ptr,4]=filter()
dbf_array[dbf_ptr,5]=relation(1)
dbf_array[dbf_ptr,6]=get_reltype(1)
dbf_array[dbf_ptr,7]=target(1)
dbf_array[dbf_ptr,8]=get_relord(1)
dbf_array[dbf_ptr,9]=relation(2)
dbf_array[dbf_ptr,10]=get_reltype(2)
dbf_array[dbf_ptr,11]=target(2)
dbf_array[dbf_ptr,12]=get_relord(2)
dbf_array[dbf_ptr,13]=relation(3)
dbf_array[dbf_ptr,14]=get_reltype(3)
dbf_array[dbf_ptr,15]=target(3)
dbf_array[dbf_ptr,16]=get_relord(3)
dbf_array[dbf_ptr,17]=relation(4)
dbf_array[dbf_ptr,18]=get_reltype(4)
dbf_array[dbf_ptr,19]=target(4)
dbf_array[dbf_ptr,20]=get_relord(4)
dbf_array[dbf_ptr,21]=relation(5)
dbf_array[dbf_ptr,22]=get_reltype(5)
dbf_array[dbf_ptr,23]=target(5)
dbf_array[dbf_ptr,24]=get_relord(5)

RETURN


PROCEDURE pop_dbf
PARAMETERS als_name
PRIVATE skip_txt,filt_expr,i,tmp_ptr

IF dbf_ptr=0
	RETURN
ENDIF

skip_txt=""
filt_expr=""
tmp_ptr=0

IF type("als_name")="C"

	FOR i=1 TO alen(dbf_array,1)
		IF dbf_array[i,1]=als_name
			tmp_ptr=i
		ENDIF
	NEXT

ELSE
	tmp_ptr=dbf_ptr
ENDIF

IF tmp_ptr=0
	RETURN
ENDIF

IF !empty(dbf_array[tmp_ptr,1])
	SELECT (dbf_array[tmp_ptr,1])
ELSE
	dbf_ptr=dbf_ptr-1

	IF dbf_ptr != 0
		DECLARE dbf_array[dbf_ptr,24]
	ENDIF

	RETURN
ENDIF

DO clear_dbf WITH (dbf_array[tmp_ptr,1])

IF !empty(dbf_array[tmp_ptr,2])
	SET ORDER TO TAG (dbf_array[tmp_ptr,2])
ENDIF

IF !empty(dbf_array[tmp_ptr,5])
	DO set_relation WITH (dbf_array[tmp_ptr,1]),(dbf_array[tmp_ptr,5]),(dbf_array[tmp_ptr,7]),(dbf_array[tmp_ptr,8])
	IF dbf_array[tmp_ptr,6]="M"
		IF !empty(skip_txt)
			skip_txt=skip_txt+", "+dbf_array[tmp_ptr,7]
		ELSE
			skip_txt=dbf_array[tmp_ptr,7]
		ENDIF
	ENDIF
ENDIF

IF !empty(dbf_array[tmp_ptr,9])
	DO set_relation WITH (dbf_array[tmp_ptr,1]),(dbf_array[tmp_ptr,9]),(dbf_array[tmp_ptr,11]),(dbf_array[tmp_ptr,12])
	IF dbf_array[tmp_ptr,10]="M"
		IF !empty(skip_txt)
			skip_txt=skip_txt+", "+dbf_array[tmp_ptr,11]
		ELSE
			skip_txt=dbf_array[tmp_ptr,11]
		ENDIF
	ENDIF
ENDIF

IF !empty(dbf_array[tmp_ptr,13])
	DO set_relation WITH (dbf_array[tmp_ptr,1]),(dbf_array[tmp_ptr,13]),(dbf_array[tmp_ptr,15]),(dbf_array[tmp_ptr,16])
	IF dbf_array[tmp_ptr,14]="M"
		IF !empty(skip_txt)
			skip_txt=skip_txt+", "+dbf_array[tmp_ptr,15]
		ELSE
			skip_txt=dbf_array[tmp_ptr,15]
		ENDIF
	ENDIF
ENDIF

IF !empty(dbf_array[tmp_ptr,17])
	DO set_relation WITH (dbf_array[tmp_ptr,1]),(dbf_array[tmp_ptr,17]),(dbf_array[tmp_ptr,19]),(dbf_array[tmp_ptr,20])
	IF dbf_array[tmp_ptr,18]="M"
		IF !empty(skip_txt)
			skip_txt=skip_txt+", "+dbf_array[tmp_ptr,19]
		ELSE
			skip_txt=dbf_array[tmp_ptr,19]
		ENDIF
	ENDIF
ENDIF

IF !empty(dbf_array[tmp_ptr,21])
	DO set_relation WITH (dbf_array[tmp_ptr,1]),(dbf_array[tmp_ptr,21]),(dbf_array[tmp_ptr,23]),(dbf_array[tmp_ptr,24])
	IF dbf_array[tmp_ptr,22]="M"
		IF !empty(skip_txt)
			skip_txt=skip_txt+", "+dbf_array[tmp_ptr,23]
		ELSE
			skip_txt=dbf_array[tmp_ptr,23]
		ENDIF
	ENDIF
ENDIF

filt_expr=alltrim(dbf_array[tmp_ptr,4])

IF !empty(filt_expr)
	SET FILTER TO &filt_expr
ENDIF

IF !empty(skip_txt)
	SET SKIP TO &skip_txt
ENDIF

IF reccount(dbf_array[tmp_ptr,1]) != 0
	GOTO dbf_array[tmp_ptr,3]
ENDIF

dbf_ptr=dbf_ptr-1

IF dbf_ptr != 0
	DECLARE dbf_array[dbf_ptr,24]
ENDIF

RETURN



****************************************************************************
*																									 
*		Function:	GET_RELTYPE
*																									
*		Purpose:		Used internally by the push_dbf procedure to determine the
*						whether the specified relationship is a one-to-one 
*						relationship or a one-to-many relationship by examining the
*						set skip setting to see if the target alias of the specified
*						relationship is present.
*
*		Returns:		"O" for a one-to-one relationship OR
*						"M" for a one-to-many relationship
*
*		Notes:		You can call this function directly if you happen to know
*						the number of the relationship you need to get the type for.
*
****************************************************************************
FUNCTION get_reltype
PARAMETERS rel_num
PRIVATE skip_txt,srch_alias,result

skip_txt=set("skip")
srch_alias=target(rel_num)
result=""

IF srch_alias $ skip_txt
	result="M"
ELSE
	result="O"
ENDIF

RETURN result



****************************************************************************
*																									 
*		Function:	GET_RELORD
*																									
*		Purpose:		Used internally by the push_dbf procedure to determine the
*						order of the target alias.
*
*		Returns:		"" if there is no master tag currently specified OR
*						the master tag of the target alias
*
*		Notes:		You can call this function directly if you happen to know
*						the number of the relationship you need to get the order 
*						of the target alias for.
*
****************************************************************************
FUNCTION get_relord
PARAMETERS rel_num
PRIVATE srch_alias,result

srch_alias=target(rel_num)
result=""

IF !empty(srch_alias)
	result=order(srch_alias)
ENDIF

RETURN result



****************************************************************************
*																									 
*		Procedure:	SET_RELATION
*																									
*		Purpose:		Used internally by the pop_dbf procedure to set the
*						relationship for the specified alias.
*
*		Notes:		You can call this function directly if you happen to know
*						the relationship alias, the relationship expression itself,
*						the target alias, and the order of the target alias.
*
****************************************************************************
PROCEDURE set_relation
PARAMETERS rel_alias,rel_expr,targ_alias,targ_order
PRIVATE cur_alias

cur_alias=alias()

SELECT (rel_alias)
SET RELATION OFF INTO (targ_alias)

SELECT (targ_alias)
SET ORDER TO TAG (targ_order)

SELECT (rel_alias)
SET RELATION TO &rel_expr INTO (targ_alias) ADDITIVE

IF !empty(cur_alias)
	SELECT (cur_alias)
ENDIF

RETURN




****************************************************************************
*																									 
*		Procedure:	CLEAR_DBF
*																									
*		Purpose:		Used internally by the pop_dbf procedure to clear the
*						all of the current settings for the specified alias.
*
*		Notes:		You can call this function directly if you happen to know
*						the alias of the work area you wish to clear.
*
****************************************************************************
PROCEDURE clear_dbf
PARAMETERS alias_name
PRIVATE cur_file

cur_file=alias()

SELECT (alias_name)
SET FILTER TO
SET RELATION TO
SET ORDER TO 0

IF !empty(cur_file)
	SELECT (cur_file)
ENDIF

RETURN



****************************************************************************
*																									 
*		Procedure:	SET_REC 
*																									
*		Purpose:		This procedure is used to modify the record number of any
*						work area saved on the stack with push_dbf.  This is very
*						useful in instances such as adding a record when you want
*						to pop the settings off the stack and you want the new
*						record number restored instead of the saved record number.
*						This way you don't have to hard-code in special handling
*						at the end of an add record procedure to go to the newly-
*						added record.
*
****************************************************************************
PROCEDURE set_rec
PARAMETERS rec_num,als_name
PRIVATE i,tmp_ptr

tmp_ptr=0

IF dbf_ptr=0
	RETURN
ENDIF

IF type("als_name")="C"

	FOR i=1 TO alen(dbf_array,1)
		IF dbf_array[i,1]=als_name
			tmp_ptr=i
		ENDIF
	NEXT

ELSE
	tmp_ptr=dbf_ptr
ENDIF

IF tmp_ptr != 0
	dbf_array[tmp_ptr,3]=rec_num
ENDIF

RETURN



****************************************************************************
*																									 
*		Procedure:	SET_FILTER
*																									
*		Purpose:		This procedure is used to modify the filter expression of 
*						any work area saved on the stack with push_dbf.  This 
*						is	very useful in instances when you want to pop the 
*						settings off the stack and you want a new filter to be 
*						restored instead of the saved filter expression.  This makes
*						it very easy to change filter expressions to reflect 
*						different views of a file.
*
****************************************************************************
PROCEDURE set_filter
PARAMETERS filt_expr,als_name
PRIVATE i,tmp_ptr

tmp_ptr=0

IF dbf_ptr=0
	RETURN
ENDIF

IF type("als_name")="C"

	FOR i=1 TO alen(dbf_array,1)
		IF dbf_array[i,1]=als_name
			tmp_ptr=i
		ENDIF
	NEXT

ELSE
	tmp_ptr=dbf_ptr
ENDIF

IF tmp_ptr != 0
	dbf_array[tmp_ptr,4]=filt_expr
ENDIF

RETURN





****************************************************************************
*																									 
*		Procedure:	GET_ALS
*																									
*		Purpose:		This procedure is used to retrieve the alias of the last
*						work area saved on the stack with push_dbf.  This is very
*						very useful in instances when you wish to re-select the
*						previously saved work area without restoring all of the 
*					 	work area settings such as relationships, filters, etc.
*
****************************************************************************
FUNCTION get_als
PRIVATE result

IF dbf_ptr=0
	result=""
ELSE
	result=dbf_array[dbf_ptr,1]
ENDIF

RETURN result






